/*
 * args.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * Argument processing functions used in pTools
 * implementation.  The three main interface functions
 * are:
 *
 * ptools_arg_setup: Parses arguments from command line and
 *   passes them to the function given as fptr.  Handles creation
 *   of internal lists of PID's and setting up libproc data structures.
 *
 * ptools_arg_getnext: Returns a proc_t structure of the next PID to
 *   be processed.
 *
 * ptools_arg_finish: Cleans up internal data structures and handles
 *   cleanup of libproc data structures.
 *
 * Compile:
 *   gcc -c args.c
 *
 * Options:
 *   If _FORCE_POSIX_ is defined arguments are processed in a POSIX way by
 *      getopt()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <proc/readproc.h>

#include <ptools-args.h>

#define TRUE  1
#define FALSE 0

#define BUFLN 1024

/* Used for Error reporting */
int ptools_errno;
char *ptools_arg_str;

/* List of PIDs passed to openproc() */
static pid_t *PIDS;

typedef struct pid_list_s {
  pid_t dataN; /* This is an int in LINUX..so neg values are ok */
  int visited;
  long int dataS_len;
  char *dataS;
  struct pid_list_s *next;
} pidlist;

/* single-linked list of PID's.  If (->dataN < 0) this argument is not
 * valid and an error should be reported.
 */
static struct ArgData {
  PROCTAB *PT;
  pidlist *head;
} proc_ents;

/* Linked-list of buffered proc_t structures */
static struct proc_list {
  proc_t *data;
  struct proc_list *next;
} *plist;


/*
 * int AttachPID(pidlist *):
 *   Attach p onto the list of arguments maintained
 * in proc_ents.  Return 0 on success, negative value
 * on error.
 */
int AttachPID(pidlist *p) {
  pidlist *cur;

  if (!proc_ents.head) {
    proc_ents.head = p;
    return 0;
  }

  cur = proc_ents.head;
  while (cur->next) { cur = cur->next; }
  
  cur->next = p;

  return 0;
}

/*
 * int AddPID(pid_t *, const char *):
 *   Validate the argument passed and attach it
 * onto the list of pids found in the proc_ents
 * structure.  The converted value is also stored in
 * dst.  Return 0 on success.  Set ptools_errno and
 * and return -1 on error.
 */
int AddPID(pid_t *dst, const char *pid) {
  pidlist *ptr;
  long int tmp = 0;
  int cnt;

  ptools_errno = NO_ERROR; /* Reset ptools_errno */

  /* Allocate space for the list object and to maintain the argument
   * passed to this function
   */
  ptr = (pidlist *) malloc(sizeof(pidlist));
  if (!ptr) { ptools_errno = MEM_ERR; return -1; }
  memset(ptr, 0, sizeof(pidlist));

  ptr->dataS_len = strlen(pid) + 1;

  ptr->dataS = (char *)malloc(sizeof(char) * ptr->dataS_len);
  if (!(ptr->dataS)) { ptools_errno = MEM_ERR; return -1; }
  memset(ptr->dataS, 0, (ptr->dataS_len * sizeof(char)));
  
  /* Is PID Valid */
  for (cnt = 0; cnt < strlen(pid); cnt++) {
    if (!isdigit(pid[cnt])) {
      tmp = INVAL_PID; break;
    }
  }

  if (!tmp) {
    errno = 0;
    tmp = strtol(pid, (char **)NULL, 10);
    if (errno == ERANGE) {
      errno = 0;
      fprintf(stderr, "Possible Overflow/Underflow error, using %ld for %s\n",
	      tmp, pid);
    }
  }

  /* Fill in list member structure */
  ptr->dataN = (pid_t)tmp;
  ptr->visited = FALSE;
  strncpy(ptr->dataS, pid, ptr->dataS_len);
  ptr->next = NULL;

  (*dst) = ptr->dataN;
  /* Attach to list and return */
  return AttachPID(ptr);
}


/*
 * void Dump_Arg_List(void):
 *    Dump the data portion of proc_ents.head
 * to stderr.  No return value.
 */
void Dump_Arg_List(void) {

  pidlist *cur;

  cur = proc_ents.head;

  while (cur) {
    fprintf(stderr, "PID: %s(%d)\n", cur->dataS, cur->dataN);
    cur = cur->next;
  }
}

/*
 * void ptools_arg_finish(void):
 *   Cleanup the memory used maintaining the linked-list
 * of PID's.  No return value.
 */
void ptools_arg_finish(void) {

  pidlist *cur, *tmp;

  if (PIDS)
    free(PIDS);  /* make sure we free the list of pids we
		  * passed to openproc()
		  */
  closeproc(proc_ents.PT);
  cur = proc_ents.head;
  
  while (cur) {
    tmp = cur;
    cur = cur->next;

    tmp->next = NULL;
    free(tmp->dataS);
    free(tmp);
  }
}


/*
 * int ptools_arg_setup(int, const char **, 
 *                  const char *, (int *)(char)):
 *
 * Initializes internal data structures for all future calls
 *  to ptools_arg_getnext().  Parses argument from command line
 *  by using getopt() and any valid argument is sent to fptr.
 *  All other non-arguments are taken as PID values and stored internally.
 *  These PID's are sent to the libproc setup function so that libproc
 *  can handle returning the proc related information.
 *
 *  Return -1 on error and sets ptools_errno to an appropriate
 * error value.  Returns 0 on success.
 */
int ptools_arg_setup(int argcnt, char * const args[], 
		     char * const opts, void *fptr(char)) {
  int cnt, ret, num_pids, pidcntr;
  extern char *optarg;
  extern int optind, opterr, optopt;
  char *options, c;

  
  /* Dump args */
  /* fprintf(stderr, "DEBUG:\tname = %s\n\targcnt = %d\n\targs[0] = %s\n",
   * name, argcnt, args[0]);
   */

  /* Initialize global data values */
  PIDS = NULL;
  ptools_errno = NO_ERROR; /* Reset ptools_errno */
  ptools_arg_str = NULL;
  plist = NULL;
  proc_ents.PT = NULL;
  proc_ents.head = NULL;

  optind = 1; /* Just incase we don't do a call to getopt */

  if ( (argcnt < 0) || (!args) ) {
    ptools_errno = INVAL_ARG;
    return -1;
  }

  /* Look at options passed */
  if (opts && fptr) {
#ifdef _FORCE_POSIX_
    if (opts[0] != '+') {
      options = (char *)malloc(sizeof(char) * (strlen(opts) + 2));
      if (!options) { ptools_errno = MEM_ERR; return -1; }
      
      options[0] = '+';
      strncpy(options+1, opts, strlen(opts));
      /*
       * fprintf(stderr, "DEBUG: arg_setup(), options: %s\n", options);
       */
    }
#else
    options = opts;
#endif
    
    while(1) {
      c = getopt(argcnt, args, options);
      
      if (c == -1) break;
      
      if (c == '?') { ptools_errno = OPT_ERR; return -1; }
      if (c == ':') { ptools_errno = NO_PARAM; return -1; }
      
      fptr(c);
    }
  } /* if (opts) */

  /*
    fprintf(stderr, "DEBUG: Allocating room for %d pid_t's\n", 
    argcnt-optind+1);
  */

  /* Allocate space for array of pid_t that is passed to openproc() */
  num_pids = argcnt-optind+1;
  PIDS = (pid_t *)malloc(sizeof(pid_t) * num_pids);
  if (!PIDS) { ptools_errno = MEM_ERR; return -1; }
  memset(PIDS, 0, (num_pids * sizeof(pid_t)));


  /* We keep our own counter for the pids array */
  pidcntr = 0;

  /* Build a list of all non-processed args */
  for (cnt = optind; cnt < argcnt; cnt++) {
    /*
      fprintf(stderr, "DEBUG: Processing %s, pidcntr: %d\n", args[cnt],
      pidcntr);
    */

    ret = AddPID(PIDS+pidcntr, args[cnt]);
    /* Advance if we assigned a valid PID during the last AddPID() *
     * otherwise we reuse the same array slot (This prevents us    *
     * from passing illegal values to openproc() below             *
     */
    if (*(PIDS+pidcntr) > 0) pidcntr++;
    else *(PIDS+pidcntr) = 0;

    if (ret < 0) {
      /* AddPID() Error */
      fprintf(stderr, "AddPID() error: %d returned\n", ret);
      free(PIDS);
      return ret;
    }
  }

  /*
    fprintf(stderr, "\nDumping Internal Argument List\n");
    Dump_Arg_List();
    fprintf(stderr, "\nList of PID's being sent to openproc() is:\n");
    for (cnt = 0; cnt < num_pids; cnt++)
    fprintf(stderr, "%d\n", PIDS[cnt]);
  */
  
  /* Call openproc() */
  proc_ents.PT = openproc(PROC_FILLBUG|PROC_PID, PIDS);

  if (!proc_ents.PT) { ptools_errno = LIBPROC_ERR; free(PIDS); return -1; }

  return 0;
}


/*
 * int BufferProcT(proc_t *):
 *   Handles buffering of proc_t items if libproc
 * returns a PID we are not expecting.
 *
 * Returns 0 on success, on error ptools_errno is set and -1
 * is returned.
 */
int BufferProcT(proc_t *p) {
  struct proc_list *cur, *tmp;

  ptools_errno = NO_ERROR; /* Reset ptools_errno */
  /* Build struct to store p in */
  tmp = (struct proc_list *)malloc(sizeof(struct proc_list));
  if (!tmp) { ptools_errno = MEM_ERR; return -1; }
  memset(tmp, 0, sizeof(struct proc_list));
    
  tmp->data = p;
  tmp->next = NULL;

  /* start list if plist is NULL */
  if (!plist) {
    plist = tmp;
    return 0;
  }

  cur = plist;

  while (cur->next) cur = cur->next;

  cur->next = tmp;
  return 0;
}


/*
 * int Visited(pid_t):
 *   See if p is in our list of PID's to process
 * and return the visited status when found.  If p
 * is not in our list of PID's return -1.
 */
int Visited(pid_t p) {
  pidlist *cur;

  cur = proc_ents.head;

  while (cur) {
    if (cur->dataN == p)
      return cur->visited;

    cur = cur->next;
  }

  return -1;
}


/*
 * proc_t *LookInBuffer(pid_t):
 *    Looks for a PID in the internal buffer, returning
 * a pointer to the proc_t structure if it is.  NULL is
 * returned if the PID is not in our buffer.
 */
proc_t *LookInBuffer(pid_t p) {

  struct proc_list *cur, *prev;
  proc_t *ret;


  prev = NULL;
  cur = plist;
  
  while (cur) {
    if (cur->data->pid == p) {
      /* found a match */
      ret = cur->data;
      if (cur == plist) plist = cur->next;
      else prev->next = cur->next;
      free(cur);
      return ret;
    } 

    /* move down the list */
    prev = cur;
    cur = cur->next;
  }
  
  return NULL;

}


/*
 * proc_t *ptools_arg_getnext(void):
 *  Returns a pointer to a valid proc_t entry, or
 * NULL on error.
 *
 *  NOTE: It is the callers responsibility to free()
 * the memory region used by the proc_t * by calling
 * freeproc(proc_t *)
 * 
 * 	This function is non-reentrant.
 */
proc_t *ptools_arg_getnext(void) {

  pidlist *cur = NULL;
  proc_t *ret = NULL;

  ptools_errno = NO_ERROR; /* Reset ptools_errno */
  ptools_arg_str = NULL;

  cur = proc_ents.head;

  /* Search for a PID that hasn't been processed yet */
  while (cur) {
    if (!cur->visited) {
      cur->visited = TRUE;
      ptools_arg_str = cur->dataS;

      /* Return NULL if this argument appears next and was invalid */
      if (cur->dataN == INVAL_PID) {
	ptools_errno = INVAL_PID;
	return NULL;
      }
      
      if (plist) { /* Search the buffer first */
	ret = LookInBuffer(cur->dataN);
	if (ret) return ret;
      }


    GetAnotherEntry:
      /* Get next proc_t from libproc */
      ret = readproc(proc_ents.PT, NULL);

      if (!ret) { /* Reached the end of the PID list */
	/*
	  ptools_errno = LIBPROC_ERR;
	*/
	ptools_errno = INVAL_PID;
 	return NULL;
      }

      /* If not what was expected buffer it if we havn't processed
       * it already
       */
      if (cur->dataN != ret->pid) {
	int tmp;
	tmp = Visited(ret->pid);
	if (tmp == TRUE) goto GetAnotherEntry;
	else if (tmp == -1) return ret; /* We don't know about this one
					 * but we return it anyway.
					 * This SHOULD never happen!!!
					 */
	else
	  BufferProcT(ret); goto GetAnotherEntry;
      }
      else
	return ret;
    } /* if (!cur->visited) */
    cur = cur->next;
  }

  return NULL;
}
