/*
 * pid.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * Implementation of functions that process PID related
 * requests in /proc.
 *
 * Compile:
 * gcc -c pid.c
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>


#define BUFLEN 512
#define ROOT    0   /* Root uid */


/*
 * ValidPid(pid_t):
 *    See if (pid) is a valid proccess
 * on the system
 *
 * Return 0 if valid, -1 otherwise
 */
int ValidPid(pid_t pid) {
  struct stat stats;
  char buf[BUFLEN];

  snprintf(buf, BUFLEN, "/proc/%d", pid);
  if (stat(buf, &stats) < 0) return -1;

  return 0;

}


/*
 * AllowExamine(pid_t):
 *    See if we should allow (pid) to be
 * examined.  This is a security function.
 *
 * NOTE:  This may not be used in any of the
 * ptools but has been left in just incase.
 *
 * Return 0 on success (allowed), -1 on error,
 * 1 if access is not allowed
 */
int AllowExamine(pid_t pid) {

  struct stat stats;
  char buf[BUFLEN];
  uid_t myuid;

  myuid = geteuid();

  if (myuid == ROOT) return 0; /* Root can examine anything */

  snprintf(buf, BUFLEN, "/proc/%d", pid);
  if (stat(buf, &stats) < 0) return -1; 

  /* DEBUG
     printf("Owner UID: %d\n", stats.st_uid);
  */
  
  if (myuid == stats.st_uid) return 0;
  return 1;
}


/*
 * OpenPid(pid_t, char *):
 *   Open /proc/pid/fname and return the file descriptor, the caller must
 * close the fd when finished with it.
 *
 * Return a valid (ie > 0) file descriptor on success or -1 on error.  errno
 * will be set by open(2) if an error occurs.
 */
int OpenPid(pid_t pid, char *fname) {

  int fd;
  char buf[BUFLEN];

  snprintf(buf, BUFLEN, "/proc/%d/%s", pid, fname);

  fd = open(buf, O_RDONLY|O_EXCL);

  /* Try open again but not exclusive (Follow from truss of Sun pcred
   * routine)
   */
  if (fd < 0) {
    fd = open(buf, O_RDONLY);
  }
  
  return fd;

}



/*
 * fOpenPid(pid_t, char *):
 *   Open /proc/pid/fname and return the FILE pointer.  The caller must close
 * the FILE object with fclose(2) when finished.
 *
 * Return a valid (ie != NULL) FILE pointer on success or NULL on error.  errno
 * will be set by fopen(3) if an error occurs.
 */
FILE *fOpenPid(pid_t pid, char *fname) {

  FILE *fd;
  char buf[BUFLEN];

  snprintf(buf, BUFLEN, "/proc/%d/%s", pid, fname);

  fd = fopen(buf, "r");

  return fd;

}
