/*
 * pcred.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * PTools Process Credential Reporting Tool
 *
 * Compile:
 *   gcc -o pcred pcred.c -lptools
 *
 * Usage:
 *   pcred { pid } ...
 *
 * Reference (given as MARKER:LineNumber in source code)
 *    ARR       linux/fs/proc/array.c
 */

#include <config.h>
#include <SMOLabs_string.h>



#define GRPSEP   " " /* Space seperated list of groups (ARR:172) */
#define GRPSEP_C ' '
/********************************
 * Global data
 ********************************/
int *grps, numgrps;

/*
 * ParseGroups(FILE *):
 *   Parse the Groups information from the
 * file pointed to by fd
 */
void ParseGroups(FILE *fd) {
  char buf[BUFLN];

  while (!feof(fd)) {
    if (!fgets(buf, BUFLN, fd))
      break;
    
    chomp(buf);

    if (!strncmp(buf, "Groups", 6)) {
      char *grouplist, *tok;
      int cnt;

      numgrps = 0;  /* should already be initialized */
      grps = NULL;

      grouplist = strdup(buf+8);

      if (grouplist) {
	chomp_space(grouplist); /* remove trailing spaces */

	/* Break-up buf to store gids in grps array */
	for (cnt = 0; cnt < strlen(grouplist); cnt++) {
	  if (grouplist[cnt] == GRPSEP_C) {
	    numgrps++;
	  }
	}
	numgrps++;
	
	grps = (int *)malloc(numgrps * sizeof(int));
	if (!grps) { free(grouplist); numgrps = 0; grps = NULL; return; }

	cnt = 0;
	tok = strtok(grouplist, GRPSEP);
	while ( tok && (cnt < numgrps) ) {
	  grps[cnt] = atoi(tok);
	  cnt++;
	  tok = strtok(NULL, GRPSEP);
	}
	
	free(grouplist);
      }
    } /* else if */
    
  }
  return;
}


/*
 * PrintCred(proc_t):
 *   Print process credentials
 */
void PrintCred(proc_t *p) {
  
  printf("%d:\t", p->pid);

  if ( (p->euid == p->ruid) && 
       (p->ruid == p->suid) && 
       (p->suid == p->fuid) )
    printf("e/r/s/fsuid=%d", p->euid);
  else
    printf("euid=%d ruid=%d suid=%d fsuid=%d", 
	   p->euid, p->ruid, p->suid, p->fuid);

  if ( (p->egid == p->rgid) && 
       (p->rgid == p->sgid) && 
       (p->sgid == p->fgid) )
    printf("  e/r/s/fsgid=%d", p->egid);
  else
    printf("  egid=%d rgid=%d sgid=%d fsgid=%d", 
	   p->egid, p->rgid, p->sgid, p->fgid);

  if (numgrps > 1) {
    int cnt;
    printf("\n\tgroups:");
    for (cnt = 0; cnt < numgrps; cnt++)
      printf(" %d", grps[cnt]);
  }
  printf("\n");
}


/*
 * ResetGlobalData:
 *   Initilization function
 */
void ResetGlobalData(void) {

  numgrps = 0;
  free(grps);
  grps = NULL;
}

void PrintUsage(char *prog) {

  fprintf(stderr, "usage: %s { pid } ...\n  (report process credentials)\n",
	 prog);
  exit(-1);
}


int main(int argc, char **argv) {

  int ret;
  FILE *statfd;
  proc_t *curproc = NULL;

  if (argc < 2) PrintUsage(argv[0]);


  /* Initialize ptools_arg library routines */
  ret = ptools_arg_setup(argc, argv, NULL, NULL);

  if (ret < 0) {
    fprintf(stderr, "Error, ptools_arg_setup(): %d\n", ptools_errno);
    return -1;
  }

  /* Loop through all pid's given on command line */
  while (1) {

    ResetGlobalData();

    curproc = ptools_arg_getnext();

    if (!curproc) {
      if (ptools_errno == 0) break;

      else if (ptools_errno == INVAL_PID)
	fprintf(stderr, "%s: cannot examine %s: no such process\n",
		argv[0], ptools_arg_str);
      else
	fprintf(stderr, "Error: %d with arg: %s\n", 
		ptools_errno, ptools_arg_str);

      continue;
    }

    /* Open status file of cur_pid */
    statfd = fOpenPid(curproc->pid, "status");
    if (!statfd) {
      printf("%s: cannot examine %d: permission denied\n",
	     argv[0], curproc->pid);
      continue;
    }

    ParseGroups(statfd);
    
    fclose(statfd);

    PrintCred(curproc);

    freeproc(curproc);
    curproc = NULL;
  } /* for() */

  ptools_arg_finish();

  return 0;
}
