/*
 * pflags.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * pTools: Process Information Utilities
 * 
 * Reports various flags and information for a process
 *
 * Compile:
 *    gcc -o -I../../include -I. -I.. -I../../lib -I../../libsl pflags pflags.c -lproc -lptools
 *
 * Usage:
 *    pflags <pid> ...
 */

#include <config.h>
#include <sig_list.h>
#include <syscall_list.h>
#include <pflags.h>

char *flags2str(unsigned long flags) {
  static char ret[BUFLN];
  char *p = ret;

  if (!flags) return "0";
  
  *p = '\0';

  if (flags & PF_ALIGNWARN)
    strcat(p, "|PF_ALIGNWARN");
  if (flags & PF_STARTING)
    strcat(p, "|PF_STARTING");
  if (flags & PF_EXITING)
    strcat(p, "|PF_EXITING");
  if (flags & PF_FORKNOEXEC)
    strcat(p, "|PF_FORKNOEXEC");
  if (flags & PF_SUPERPRIV)
    strcat(p, "|PF_SUPERPRIV");
  if (flags & PF_DUMPCORE)
    strcat(p, "|PF_DUMPCORE");
  if (flags & PF_SIGNALED)
    strcat(p, "|PF_SIGNALED");
  if (flags & PF_MEMALLOC)
    strcat(p, "|PF_MEMALLOC");
  if (flags & PF_USEDFPU)
    strcat(p, "|PF_USEDFPU");

  if (*p == '|') p++;

  return p;
}


/* We do the open()/read() first becuase if it fails we
 * shouldn't return any information about the process.
 * This is close to the way Solaris does it and could
 * be changed
 */
int do_pflags(proc_t *p) {
  int fd;
  char buf[BUFLN];
  unsigned long long sigpend;
  u_int32_t pend1, pend2;

  struct ptools_struct pt_info;


  /* Solaris has multiple LWP's but we don't (really),
   * this support will be added later
   */
  snprintf(buf, BUFLN, "/proc/%d/stack", p->pid);
  fd = open(buf, O_RDONLY);
  if (fd < 0) {
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "Error on open() of %s\n", buf);
#endif
    return -1;
  }

  if (read(fd, &pt_info, sizeof(struct ptools_struct)) <
      sizeof(struct ptools_struct)) {
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "Error on read() of %s\n", buf);
#endif
    return -1;
  }



  printf("%d:\t%s\n", p->pid, *(p->cmdline));
  /* NOTE: This is hard-coded for the i386 architecture
   * We should change this by accessing the binary with the
   * elf libraries if possible
   */
  printf("\tdata model = _ILP32");

  if (p->flags) {
    printf("  flags = %s", flags2str(p->flags));
  }
  printf("\n");


  sigpend = ConvertMask(p->signal);
  if (sigpend < 0) {
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "ConvertMask(%s) failed.\n", p->signal);
#endif
    return -1;
  }

  pend1 = *((u_int32_t*)(&sigpend));
  pend2 = *( ((u_int32_t*)&sigpend) + 1);

  /* Gotta switch the order of printing due to endian diff
   * between SPARC and Intel
   */
  if (pend1 || pend2)
    printf("  sigpend = 0x%.8x,0x%.8x\n", pend2, pend1);

  /* We would have individual flags if we had LWP's
   * printf("  /1:\tflags: = %s", );
   */
  printf("  /1:");
  if (pt_info.pr_syscall >= FIRST_SYSCALL) {
    int n;
    /* skip over sys_ in name returned from syscallnum2str() */
    printf("\t[ %s(", syscallnum2str(pt_info.pr_syscall, buf, BUFLN)+4);

    for(n = 0; n < syscall_numargs(pt_info.pr_syscall); n++) {
      if (n) printf(",");
      printf("0x%lx", pt_info.pr_sysarg[n]);
    }

    printf(") ]");
  }
  printf("\n");

  

  return 0;
}



void PrintUsage(const char *name) {
  const char *ptr;

  if ( (ptr = strrchr(name, '/')) != NULL)
    ptr += 1;
  else
    ptr = name;

  fprintf(stderr, "usage:  %s pid ...\n", ptr);
  fprintf(stderr, "  (report process status flags)\n");
  /*
    fprintf(stderr, "  -r : report registers\n");
  */
  
  exit(-1);
}


int main(int argc, char **argv) {

  proc_t *curproc;

  if (argc < 2) PrintUsage(argv[0]);

  if (ptools_arg_setup(argc, argv, NULL, NULL) < 0) {
    fprintf(stderr, "Error, ptools_arg_setup(): %d\n",
	    ptools_errno);
    return -1;
  }

  while (1) {
    curproc = ptools_arg_getnext();
    if (!curproc) {
      if (ptools_errno == 0) break;

      else if (ptools_errno == INVAL_PID)
        fprintf(stderr, "%s: cannot examine %s: no such process\n",
                argv[0], ptools_arg_str);
      else
        fprintf(stderr, "Error: %d with arg: %s\n", 
                ptools_errno, ptools_arg_str);

      continue;
    }

    if (do_pflags(curproc) < 0)
      printf("Error processing PID %d, skipping.\n", curproc->pid);

    freeproc(curproc);
    curproc = NULL;
  } /* while() */

  ptools_arg_finish();

  return 0;
}
