/*
 * pldd.c
 * Stphen Oberther
 * oberther@cs.fsu.edu
 *
 * pTools Process Library Reporting Tool
 *
 * Reports the dynamic libraries linked to a process
 *
 * Usage:
 *   pldd [ pid ] . . .
 *
 * NOTES:
 *   Check on how -F flag in Solaris version is implemented,
 * has to do with process locking I think.
 *
 */

#include <config.h>

#include <ptools_maps.h>


extern FILE *yyin;
int yylex(void);

extern struct maps_ent *head;

/* Linked-List of library names */
struct liblist {
  char *name;
  struct liblist *next;
};



/*
 * void CleanupMapsList(struct maps_ent *):
 *   Reclaims dynamic memory of the single-linked
 * list of map file entries parsed by lex
 */
void CleanupMapsList(struct maps_ent *p) {
   struct maps_ent *cur;

   cur = p;

   while (cur) {
     struct maps_ent *tmp;

     tmp = cur;
     cur = cur->next;
     free(tmp);
   }

}


/*
 * void DumpMapsList(struct maps_ent *):
 *   Prints out maps_ent structures to stdout
 */
void DumpMapsList(struct maps_ent *p) {

  if (p->next) DumpMapsList(p->next);

  printf("%08lx-%08lx %s %08lx %02d:%02d %lu\t%s\n", p->start, p->end, 
	 p->perms, p->offset, p->maj, p->min, p->inode, p->path);
}


/*
 * void FreeLibrarList(struct liblist *):
 *   Reclaims dynamic memory used in creating
 * the linked list of library names. The name data
 * element is not reclaimed because it points to the
 * static array "path" in the maps_ent structure.
 */
void FreeLibraryList(struct liblist *p) {
  struct liblist *cur;

  cur = p;
  while (cur) {
    struct liblist *tmp;

    tmp = cur;
    cur = cur->next;
    free(tmp);
  }

}


/*
 * void DumpLibraryList(struct liblist *):
 *   Prints the linked list of library
 * names to stdout.
 */
void DumpLibraryList(struct liblist *p) {
  
  struct liblist *cur;

  cur = p;

  while (cur) {
    printf("%s\n", cur->name);
    cur = cur->next;
  }

}

/*
 * IsShared(const char *):
 *   Return true if the shared flag is set
 * in p, false if not
 */
int IsShared(const char *p) {

  if (p[3] == 's') return TRUE;

  return FALSE;

}


/*
 * BuildLibraryList(struct maps_ent *):
 *
 *   Build a unique list of libraries, excluding shared
 * memory regions, from the mapping information in p
 * Return the list, or NULL on error
 */
struct liblist *BuildLibraryList(struct maps_ent *p) {

  struct liblist *ret, *ptr;
  struct maps_ent *cur;

  ret = NULL;
  cur = p;

  while (cur) {
    if (*cur->path) {

      /*
	printf("%s\t", cur->path);
      */

      if (IsShared(cur->perms)) {
	cur = cur->next; continue;
      }

      if (ret) {
	/* Based on 2.4.x, x <= 13, maps file is layed out so that 
	 * only need to compare current against most recent added to
	 * list */
	if ( (strlen(ret->name) == strlen(cur->path)) &&
	     (strncmp(ret->name, cur->path, strlen(ret->name)) == 0) ) {
	  /* Already added cur->path to library list */

	  /* 
	     printf("SKIPPED\n");
	  */

	  cur = cur->next;
	  continue;
	}
      }
      
      ptr = (struct liblist *)malloc(sizeof(struct liblist));
      if (!ptr) {
	fprintf(stderr, "Memory allocation error\n");
 	FreeLibraryList(ret);
	return NULL;
      }
      /* Latch ptr onto list */
      ptr->name = cur->path;
      ptr->next = ret;
      ret = ptr;

      /*
	printf("ADDED\n");
      */

    }
    cur = cur->next;
  }

  return ret;
}


/*
 * void PrintUsage(const char *):
 *   Prints the program usage.
 */
void PrintUsage(const char *name) {
  fprintf(stderr, 
	  "usage:  %s { pid } ...\n  (reports process dynamic libraries)\n", 
	  name);
  exit(-1);
}



/* No Comment */
int main(int argc, char **argv) {
  int ret;
  char buf[BUFLN];
  FILE *ifile;
  
  struct liblist *libhead = NULL;

  proc_t *curproc;


  if (argc < 2) PrintUsage(argv[0]);

  ret = ptools_arg_setup(argc, argv, NULL, NULL);

  if (ret < 0) {
    fprintf(stderr, "Error, ptools_arg_setup(): %d\n", ptools_errno);
    return -1;
  }

  /* Go through each returned value and parse
   * the maps file for each pid returned from readproc
   */
  while (1) {
    curproc = ptools_arg_getnext();
    if (!curproc) {
      if (ptools_errno == 0) break;

      else if (ptools_errno == INVAL_PID)
        fprintf(stderr, "%s: cannot examine %s: no such process\n",
                argv[0], ptools_arg_str);
      else
        fprintf(stderr, "Error: %d with arg: %s\n", 
                ptools_errno, ptools_arg_str);

      continue;
    }

    printf("%d:\t", curproc->pid);

    snprintf(buf, BUFLN, "/proc/%d/maps", curproc->pid);

    /* Open the input file */
    ifile = fopen(buf, "r");
    if (!ifile) {
      fprintf(stderr, "Error opening: %s\n", buf);
      continue;
    }

    /* Parse the maps file */
    yyin = ifile;
    yylex();
    fclose(ifile);

    /*
      printf("\n********************************\n");
      DumpMapsList(head);
      printf("\n********** Building Library List ************\n");
    */

    /* Build the list of libraries */
    libhead = BuildLibraryList(head);
    if (!libhead) {
      printf("\n");
      continue;
    }

    /* Print the list of libraries and cleanup */
    DumpLibraryList(libhead);

    CleanupMapsList(head);
    FreeLibraryList(libhead);
    head = NULL;
    libhead = NULL;
    freeproc(curproc);
    curproc = NULL;
  }

  /* Final cleanup */
  CleanupMapsList(head);
  FreeLibraryList(libhead);
  head = NULL;
  libhead = NULL;
  freeproc(curproc);
  curproc = NULL;

  ptools_arg_finish();

  return 0;
}
