/*
 * psig.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * pTools: Process Credential Reporting Tool
 *
 * Reports information regarding signal handeling for a process.
 *
 * Usage:
 *    psig <pid> ...
 */

#define _PTOOLS_DEBUG
#include <config.h>
#include <signal.h>
#include <sigact.h>
#include <sig_list.h>



char *SignalFlags(int sig, int flags, char *buf, size_t len) {

  int flagmask = (SA_ONSTACK|SA_RESTART|SA_NODEFER|SA_RESETHAND|SA_SIGINFO);

  if (sig == SIGCLD)
    flagmask |= (SA_NOCLDSTOP|SA_NOCLDWAIT);

  *buf = '\0';

  if (flags & ~flagmask) {
    snprintf(buf, len, ",0x%x,", flags & ~flagmask);
    len -= strlen(buf) - 1;
  }
  else if (flags == 0)
    return(buf);

  if (len < 1) goto out;

  if (flags & SA_RESTART) {
    strcat(buf, ",RESTART");
    len -= 8; if (len < 1) goto out;
  }
  if (flags & SA_RESETHAND) {
    strcat(buf, ",RESETHAND");
    len -= 10; if (len < 1) goto out;
  }
  if (flags & SA_ONSTACK) {
    strcat(buf, ",ONSTACK");
    len -= 8; if (len < 1) goto out;
  }
  if (flags & SA_SIGINFO) {
    strcat(buf, ",SIGINFO");
    len -= 8; if (len < 1) goto out;
  }
  if (flags & SA_NODEFER) {
    strcat(buf, ",NODEFER");
    len -= 8; if (len < 1) goto out;
  }


  if (sig == SIGCLD) {
    if (flags & SA_NOCLDWAIT) {
      strcat(buf, ",NOCLDWAIT");
      len -= 10; if (len < 1) goto out;
    }
    if (flags & SA_NOCLDSTOP) {
      strcat(buf, ",NOCLDSTOP");
      len -= 10;
    }
  } /* if (sig == SIGCLD) */

 out:

  *buf = '\t';
  return(buf);
}

/*
 * inmask(int, unsigned long long):
 *   Checks to see if sig is contained within
 * mask.  Returns 1 iff sig is found in mask.
 */
int inmask(int sig, unsigned long long mask) {

  mask = mask >> (sig - 1);
  if (0x1 & mask) return 1;

  return 0;
}


/*
 * do_sigact(proc_t):
 *
 */
int do_sigact(proc_t *p) {
  unsigned long long blockmask;
  char buf[BUFLN];
  int tmp, fd, sig;
  struct kern_sigaction signals[NSIG], *cursig;

  cursig = NULL;
  memset(signals, 0, NSIG*sizeof(struct kern_sigaction));

  /* Setup mask of blocked signals */
  if ( (blockmask = ConvertMask(p->blocked)) < 0) {
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "ConvertMask(%s) failed.\n", p->blocked);
#endif
    return -1;
  }

#ifdef _PTOOLS_DEBUG
  printf("blockmask : 0x%016Lx\n", blockmask);
#endif


  /* Open sigact file */
  snprintf(buf, BUFLN, "/proc/%d/sigact", p->pid);
  if ( (fd = open(buf, O_RDONLY)) < 0) {
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "open(%s, O_RDONLY): failed, %d\n", buf, errno);
#endif
    return -1;
  }

  tmp = NSIG * sizeof(struct kern_sigaction);
  if (read(fd, (void *)signals, tmp) != tmp) {
    close(fd);
#ifdef _PTOOLS_DEBUG
    fprintf(stderr, "read(%d, %p, %d): failed, %d\n", fd, signals, tmp, errno);
#endif
    return -1;
  }
  close(fd);


  /* Print PID and command line */
  printf("%d:\t%s\n", p->pid, p->cmdline[0]);

  /* Go through each signal */
  for(sig = 1; sig < SIGRTMIN; sig++) {
    char str[100];
    
    tmp = 0;
    cursig = &(signals[sig-1]);
    cursig->flags &= ~SA_RESTORER; /* SA_RESTORER is set by
				    * libc, during the library call
				    * so we remove it here
				    */
    printf("%s\t", signum2str(sig, buf, BUFLN)+3); /* Skip the initial 
						    * SIG */
    /* Print status */
    if (inmask(sig, blockmask))
      printf("blocked,");

    if (cursig->handler == SIG_DFL)
      printf("default");

    else if (cursig->handler == SIG_IGN)
      printf("ignored");

    else {
      printf("caught");
      tmp = 1;
    }

    /* If caught, go through and print the signals that are blocked
     * while processing the current signal */
    if (tmp) {
      int cnt;
      SignalFlags(sig, cursig->flags, str, 100);

      printf("%s", (*str != '\0')?str : "\t0");
      tmp = 0;
      for (cnt = 1; cnt < NSIG; cnt++) {
	if (inmask(cnt, cursig->mask)) {
	  printf(tmp++? "," : "\t");
	  printf("%s", signum2str(cnt, str, 100)+3);
	}

      } /* for() */
      
    }

    else if (sig == SIGCLD) {
      printf("%s", SignalFlags(sig, cursig->flags & (SA_NOCLDSTOP),
			       str, 100));
    }
    
    printf("\n");
  }
  return 0;
}




void PrintUsage(char *prog) {

  fprintf(stderr, "usage:\t%s pid ...\n", prog);
  fprintf(stderr, "  (report process signal actions)\n");

  exit(-1);
}



int main(int argc, char **argv) {
  int ret;
  proc_t *curproc = NULL;

  if (argc <= 1) PrintUsage(argv[0]);

  /* Initialize ptools_arg library routines */
  ret = ptools_arg_setup(argc, argv, NULL, NULL);

  if (ret < 0) {
    fprintf(stderr, "Error, ptools_arg_setup(): %d\n", ptools_errno);
    return -1;
  }

  /* Loop through all pid's given on command line */
  while (1) {
    curproc = ptools_arg_getnext();

    /* Deal with errors */
    if (!curproc) {
      if (ptools_errno == 0) break;
      else if (ptools_errno == INVAL_PID)
	fprintf(stderr, "%s: cannot examine %s: no such process\n",
		argv[0], ptools_arg_str);
      else
	fprintf(stderr, "Error: %d with arg: %s\n",
		ptools_errno, ptools_arg_str);

      continue;
    }

    /* Perform the actual psig operation on the current process
     * we are investigating */
    do_sigact(curproc);
    freeproc(curproc);
    curproc = NULL;
  }
  
  ptools_arg_finish();

  return 0;
}
