/*
 * ptree.c
 * Stephen Oberther
 * oberther@cs.fsu.edu
 *
 * pTools: Process Information Utilities
 *
 * Displays process relationships in a tree format
 *
 * Compile:
 *   gcc -Wall -g -o ptree ptree.c -lproc
 * Requires:
 *   libproc (Available in procps package 
 *              ftp://people.redhat.com/johnsonm/procps/)
 */

#include <config.h>

#define INIT 1

enum type { ProcessID, UserID };

enum bool_t { false = 0, true = 1};
typedef enum bool_t bool;

/* Tree node structure */
struct tnode_t {
  proc_t *pent;
  bool visited;
  struct tnode_t *l, *r;
};
typedef struct tnode_t tnode;


/* Link List node */
struct llnode_t {
  tnode *data;
  struct llnode_t *next;
};
typedef struct llnode_t llnode;


int PrintInit = 0;
const char *PROGNAME;
tnode *tree;

/*
 * tnode *MkTnode(proc_t, tnode *, tnode *):
 *   Makes a tree node and initializes appropriate
 * fields.  Returns tree node or NULL on error.
 */
tnode *MkTnode(proc_t *data, tnode *l, tnode *r) {
  tnode *ret;

  ret = (tnode *)calloc(1, sizeof(tnode));
  if (!ret)
    return NULL;

  ret->pent = data;
  ret->l = l; ret->r = r;
  ret->visited = false;

  return ret;
}

/***********************************
 * Modified from readproc.c in
 * procps 2.0.7
 ***********************************/
/* tnode *LookupPID(tnode *, int):
 *   Traverse tree 't' breadth-first looking 
 * for a process with pid p.  Return pointer
 * to tree entry or NULL on error
 */
tnode *LookupPID(tnode* t, int p) {

  tnode* tmp = NULL;

  if (!t)
    return NULL;

  /* look here/terminate recursion */
  if ( (t->pent) && (t->pent->pid == p) )
    return t;
  
  /* recurse over children */
  if ((tmp = LookupPID(t->l, p))) 
    return tmp;

  /*  recurse over siblings */

  if ((tmp = LookupPID(t->r, p)))
    return tmp;

  return NULL;
}


/*
 * tnode *MakeTree(PROCTAB *):
 *    Make a tree from the proc entries
 * found in p and return it.  Builds a 
 * Left child, Right sibling tree.
 * NULL returned on error.
 */
tnode *MakeTree(PROCTAB *p) {

  tnode *t = NULL;
  tnode *srch = NULL;
  tnode *tmp = NULL;
  proc_t *cur;

  if (!p)
    return NULL;

  cur = readproc(p, NULL);

  if (cur)   t = MkTnode(cur, NULL, NULL);
  if (!t) return NULL;

  if (cur->pid != INIT)   t = MkTnode(NULL, NULL, t);
  if (!t) return NULL;

  /* Pass one over /proc, generates initial tree */
  while ( (cur = readproc(p, NULL)) ) {

    if (cur->pid == INIT) {
      t->pent = cur;
      continue;
    }

    srch = LookupPID(t, cur->ppid);

    if (srch) {
      /* Latch child(cur) under right side of parent(srch) */
      tmp = MkTnode(cur, NULL, srch->l);
      if (!tmp) return NULL;
      srch->l = tmp;
    }

    else {
      /* Place cur on right side of tree to be scanned next */
      /* THIS SHOULD NEVER HAPPEN WITH LIBPROC */
      tmp = MkTnode(cur, NULL, t->r);
      if (!tmp) return NULL;
      t->r = tmp;
    }
    
  }

  /* Pass two: Look at tree.r only */
  /* Not needed with libproc-2.0.7, processes are returned in a sorted
   * manner, apparently by process creation time
   */

  return t;
}


/*
 * ClearVisited(tnode *t):
 *    Sets visited to false for t and all child nodes
 * (ie t->l and t->r) of t.
 */
void ClearVisited(tnode *t) {
  if (!t) return;
  
  ClearVisited(t->r);
  ClearVisited(t->l);

  t->visited = false;
}

/*
 * FreeTree(tnode *):
 *   Reclaim the memory used by t
 */
void FreeTree(tnode *t) {

  if (!t) return;

  FreeTree(t->r);
  FreeTree(t->l);

  freeproc(t->pent);
  t->r = NULL; t->l = NULL;
  free(t);
}


/*
 * PrintProc(proc_t *, int):
 *   Actually prints out a single process line
 * of a process tree.
 */
void PrintProc(proc_t *p, int t) {

  if (!p) return;

  if (t)
    printf("%*c", t, ' '); /* Print 'tab' spaces */
  
  printf("%-6d", p->pid);
  
  if (p->cmdline)
    printf("%s\n", p->cmdline[0]);
  else
    printf("%s\n", p->cmd);

}


/*
 * PrintTree(proc_t *):
 *   Prints the tree t in a manner similiar
 * to ptree on Solaris.
 */
void PrintTree(tnode *t, int tab) {

  if (!t) return;

  PrintTree(t->r, tab);

  if (t->visited == false) {
    PrintProc(t->pent, tab);
    t->visited = true;
  }

  if (t->l) {
    tab += 2;
    PrintTree(t->l, tab);
    tab -= 2;
  }

}



/*
 * PrintList(llnode *):
 *   Handles outputing the elements of 'list' in
 * a tree format
 */
void PrintList(llnode *list) {
  llnode *tmp;
  int tab;


  if (!list) return;
  /* Print out each list element and then the rest of the tree
   * from the last item
   */
  tab = 0;
  for (tmp=list; tmp; tmp=tmp->next) {
    /* continue should be just fine */
    if ( (!tmp->data) || (!tmp->data->pent) ) continue;

    if ( (tmp->data->pent->pid == 1) && (!PrintInit) ) continue;

    if (tmp->data->visited) { tab+=2; continue; }

    PrintProc(tmp->data->pent, tab);
    tmp->data->visited = true;
    tab +=2;
    
    if (!(tmp->next)) break;
  }

  if (tmp) /* Need to check incase we continue out of loop */
    PrintTree(tmp->data->l, tab);

}


/*
 * FreeList(llnode *):
 *   Cleans up the dynamic memory
 * used by 'list'
 */
void FreeList(llnode *list) {
  llnode *tmp, *holder;
  
  /* Clean up the list */
  tmp = list;
  while (tmp) {

    holder = tmp;
    tmp = tmp->next;
    holder->data = NULL;
    holder->next = NULL;
    free(holder);
  }
}

/*
 * PrintPID(tnode *, int):
 *   Print out parents and children of process 'pid'
 * Searches starting at 't'.  If 't' is NULL search
 * from global var 'tree'
 */
void PrintPID(tnode *t, int pid) {

  llnode *list = NULL;
  llnode *tmp;
  tnode *n;
  int id;

  id = pid;

  if (!t)
    t = tree;
  
  /* Build list of the nodes to be printed */
  /* Could be optimized to skip adding n if it has already
   * been visited
   */
  while ( (n = LookupPID(t, id)) ) {

    tmp = (llnode *)malloc(sizeof(llnode));
    if (!tmp) {
      printf("Internal error: malloc()\n");
      return;
    }

    tmp->next = list;
    tmp->data = n;

    list = tmp;

    if (!n->pent) break; /* can't get parent id */
    id = n->pent->ppid;
  }

  PrintList(list);
  FreeList(list);
}


/*
 * PrintUIDTree(tnode *, int, int):
 *   Prints out all the process owned by user 'id'
 */
void PrintUIDTree(tnode *t, int tab, int id) {

  if (!t) return;

  PrintUIDTree(t->r, tab, id);

  if ( (t->visited == false) && (t->pent) && (t->pent->ruid == id) ) {
    /* Build up a list of parents */
    PrintPID(NULL, t->pent->pid);
  }


  if (t->l) {
    tab += 2;
    PrintUIDTree(t->l, tab, id);
    tab -= 2;
  }

}


/*
 * PrintUID(tnode *, const char *):
 *   Print all processes owned by 'id'
 * Looks up 'id' in the passwd file and calls
 * PrintUIDTree() which does all the work
 */
void PrintUID(tnode *t, const char *id) {
  struct passwd *pwent;
  uid_t uid;

  /* Lookup id */
  pwent = getpwnam(id);
  if (!pwent) {
    printf("%s: cannot find %s passwd entry\n", PROGNAME, id);
    return;
  }
  uid = pwent->pw_uid;

  /* Print the actual processes in tree format
   * that are owned by id */
  PrintUIDTree(t, 0, (int)uid);

}

/*
 * PrintUsage():
 *   Print program usage
 */
void PrintUsage(void) {

  fprintf(stderr, "usage: %s [-a] [ {pid|user} ... ]\n", PROGNAME);
  fprintf(stderr, "  (show process trees)\n");
  fprintf(stderr, "  list can include process-ids and user names\n");
  fprintf(stderr, "  -a : include children of process 0\n");
}


/*
 * ProcessArgs(tnode *, int, char **):
 *   Processes arguments sent to the program
 */
void ProcessArgs(tnode *t, int argc, char **argv) {
  int cnt;
  char c;
  extern char *optarg;
  extern int optind, opterr, optopt;

  while (1) {
    /* '+' forces POSIX compliance */
    c = getopt(argc, argv, "+a");

    if (c == -1)
      break;

    switch(c) {
    case 'a':
      PrintInit = 1;
      break;

    case '?':
      PrintUsage();
      return;
    };

  }


  cnt = optind;  /* first pid/uid is at argv[optind] */

  if (cnt == argc) {
    if (PrintInit) PrintTree(t, 0);
    else PrintTree(t->l, 0);
  }
  else
    for (; cnt < argc; cnt++) {
      if (isdigit(argv[cnt][0]))
	PrintPID(t, atoi(argv[cnt]));
      else
	PrintUID(t, argv[cnt]);
    }

}

int main(int argc, char **argv) {

  PROCTAB *PT   = NULL;

  PROGNAME = argv[0];

  PT = openproc(PROC_FILLCMD|PROC_FILLSTAT|PROC_FILLSTATUS, 0);
  if (!PT) {
    fprintf(stderr, "Internal openproc() error\n");
    return -1;
  }

  tree = MakeTree(PT);
  
  if (!tree) {
    closeproc(PT);
    return -1;
  }

  ProcessArgs(tree, argc, argv);

  closeproc(PT);
  FreeTree(tree);
  
  return 0;
}
